/*

                                   _JNJ`                        
                               .JNMH`                           
                              JMMF`       `;.                   
                            .NMM)          `MN.                 
                            MMM)            (MML                
                           (MMM`             MMML               
                     M  I  D  N  I  G  H  T    C  o  D  E       
                           (NMMF             MHNH               
                            NMML            .MMM                
                             NMML          .NMH                 
                              4MMNL       .#F                   
                               `4HNNL_   `                      
                                   `"""`                        

                          Copyright (C) 2004-2006
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/

/* htmlnx.c :: Midnight Code HTML test and example program		*/


#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <arpa/inet.h>

#define MIDNIGHTCODE_LIST_NOLOCK
#include "libmidnightcode/libmidnightcode.h"


int
bin_to_std_err(char * value, char * abs_url, char * rel_url) {
  int url_len;
  int ret_i;

  url_len = 0;
  ret_i = 0;

  // printf("Debug URLs a: [%s]\n", value);

  url_len = strlen(value);
  if(!(midnightcode_parse_exttype(value, url_len,
                                  MIDNIGHTCODE_PARSE_EXT_TYPE_PICTURE) ||
       midnightcode_parse_exttype(value, url_len,
                                  MIDNIGHTCODE_PARSE_EXT_TYPE_VIDEO) ||
       midnightcode_parse_exttype(value, url_len,
                                  MIDNIGHTCODE_PARSE_EXT_TYPE_ARCHIVE)))
    return 0;

  ret_i = midnightcode_html_urltype(value, url_len);

  switch(ret_i) {
    case MIDNIGHTCODE_HTML_URL_TYPE_ABSOLUTE:
      fprintf(stderr, "http://%s%s\n", abs_url, value);
      break;

    case MIDNIGHTCODE_HTML_URL_TYPE_RELATIVE:
      fprintf(stderr, "http://%s/%s\n", rel_url, value);
      break;

    case MIDNIGHTCODE_HTML_URL_TYPE_FULL:
      fprintf(stderr, "http://%s\n", value);
      break;

    case MIDNIGHTCODE_HTML_URL_TYPE_URI:
      fprintf(stderr, "%s\n", value);
      break;

    default:
      return 0;

  }

  return 1;
}


int
main(int argc, char **argv) {
  int ret_i;
  char input_file_name[64] = "\0";
  char original_url[256] = "\0";
  char absolute_url[256] = "\0";
  char relative_url[256] = "\0";
  int help = 0;
  void * tmp_void_p;
  struct midnightcode_html_document_head_t * html_list_head_p;
  struct midnightcode_html_document_element_t * now_document;
  struct midnightcode_html_tag_element_t * now_tag;
  int list_num;
  int fp;
  char buffer[32000];
  int found_a;

  // Prepare switches
  struct midnightcode_switch_t switches[] = {
    { "if", "input_file", "HTML file for parsing",
      MIDNIGHTCODE_PARSE_TYPE_STRING, 64, 0, &input_file_name },
    { "ou", "orig_url", "Original URL",
      MIDNIGHTCODE_PARSE_TYPE_STRING, 256, 0, &original_url },
    { "h", "help", "Display this help/usage screen",
      MIDNIGHTCODE_PARSE_TYPE_NONE, 0, 0, &help },
    { 0 }
  };

  // Prepare HTML
  midnightcode_list_init(html_struct_head,
                         midnightcode_html_document_element_t);
  tmp_void_p = (void *)&html_struct_head;
  html_list_head_p = tmp_void_p;

  // Process switches
  if((ret_i = midnightcode_switch_process(argc, argv, switches)) != 0) {
    fprintf(stdout, "\n");
    midnightcode_logo(0,
      MIDNIGHTCODE_LICENSE_STYLE_BLOOD_MOON);
    midnightcode_license(0,
      MIDNIGHTCODE_LICENSE_STYLE_BLOOD_MOON, 
      MIDNIGHTCODE_LICENSE_TEXT_COPYRIGHT);
    fprintf(stdout, "    Error in switches, switch \"%s\" is unkown"
            " at parameter number %d.\n\n", argv[ret_i], ret_i);
    midnightcode_switch_usage(MIDNIGHTCODE_SWITCH_STYLE_SOFT,
                              argv[0], switches);
    exit(1);
  }
  if(help) {
    midnightcode_logo(0, MIDNIGHTCODE_LICENSE_STYLE_BLOOD_MOON);
  midnightcode_license(0,
    MIDNIGHTCODE_LICENSE_STYLE_BLOOD_MOON, 
    MIDNIGHTCODE_LICENSE_TEXT_COPYRIGHT);
    midnightcode_switch_usage(MIDNIGHTCODE_SWITCH_STYLE_SOFT,
                              argv[0], switches);
    exit(1);
  }

  fflush(stdout);

  // Process license and logo
  midnightcode_logo(0,
    MIDNIGHTCODE_LICENSE_STYLE_BLOOD_MOON);
  midnightcode_license(0,
    MIDNIGHTCODE_LICENSE_STYLE_BLOOD_MOON, 
    MIDNIGHTCODE_LICENSE_TEXT_ALL);


  // Program title
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_H1,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "HTML File Parser: %s", argv[0]);

  fprintf(stdout, "\n");
  fflush(stdout);

  // Validate input file
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_TEXT,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "Validating input file");
  if(!strlen(input_file_name) ||
     (access(input_file_name, F_OK) != 0)) {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_FAIL);
    fprintf(stdout, "\n");
    exit(1);
  } else {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_OKAY);
  }

  // Opening input file
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_TEXT,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "Opening input file");
  if(!(fp = open(input_file_name, O_RDONLY))) {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_FAIL);
    fprintf(stdout, "\n");
    exit(1);
  } else {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_OKAY);
  }

  // Reading input file
  memset(buffer, 0, sizeof(buffer));
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_TEXT,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "Reading input file");
  if(read(fp, buffer, sizeof(buffer)) < 1) {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_FAIL);
    fprintf(stdout, "\n");
    exit(1);
  } else {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_OKAY);
  }

  // Closing input file
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_TEXT,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "Closing input file");
  if(close(fp)) {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_FAIL);
    fprintf(stdout, "\n");
    exit(1);
  } else {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_OKAY);
  }


  // Process HTML links
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_TEXT,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "Parsing HTML content");
  if((ret_i = midnightcode_html_parse(buffer, strlen(buffer),
                           html_list_head_p)) != 0) {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_FAIL);
  } else {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_OKAY);
  }


  // Checking Base URL
  midnightcode_status_text(MIDNIGHTCODE_STATUS_TYPE_TEXT,
                           MIDNIGHTCODE_STATUS_STYLE_NORMAL,
                           "Validating base URL");
  if(!strlen(original_url) || 
     midnightcode_html_urlbase(absolute_url, original_url,
       strlen(original_url), MIDNIGHTCODE_HTML_URL_TYPE_ABSOLUTE) < 1 ||
     midnightcode_html_urlbase(relative_url, original_url,
       strlen(original_url), MIDNIGHTCODE_HTML_URL_TYPE_RELATIVE) < 1) {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_FAIL);
  } else {
    midnightcode_status_flag(MIDNIGHTCODE_STATUS_FLAG_OKAY);
  }

  fprintf(stdout, "\n");
  found_a = 0;
  midnightcode_list_for_each(now_document, html_list_head_p) {
    if(now_document) {
      if(now_document->tag) {
        list_num = 0;
        found_a = 0;
        midnightcode_list_for_each(now_tag, now_document->tag) {
          list_num++;
          if(list_num == 1) {
            if(now_tag->option) {
              if(strcasecmp(now_tag->option, "a") == 0) {
                found_a = 1;
                continue;
              }
            }
          }
          if(found_a && now_tag->option) {
            if(strcasecmp(now_tag->option, "href") == 0) {
              if(now_tag->value) {
                ret_i = bin_to_std_err(now_tag->value,
                                       absolute_url, relative_url);
              }
            }
          }
        }
      }
    }
  }

  exit(0);
}

